# SecMonitor: 一个基于eBPF技术的命令行容器行为监控工具 开发文档

Proj118: An eBPF-based Monitor for Container 基于eBPF技术的容器行为监控工具

New York University Shanghai 上海纽约大学

NYUSHTeam1 上海纽约大学1队 门秋彤(qiutong.m@nyu.edu)

指导教师: 刘古月(guyue.liu@gmail.com), 程泽睿智(chengzeruizhi@huawei.com)

---

## 项目描述

容器技术对系统资源实现了较好的隔离, 因而在云服务生产和开发环境中广泛应用. 然而, 默认的Docker环境需要赋予容器最高权限, 带来巨大的安全风险. SECCOMP是Linux内核的特性，开发者可以通过SECCOMP限制容器的行为, 而Docker提供的默认SECCOMP配置允许了200余项系统调用的权限, 其中大多数是容器正常服务所不需要的, 由此极大地引入了容器逃逸的安全风险.

eBPF是Linux内核中的一个执行引擎, 可以在沙盒环境中安全高效地在内核中执行自定义程序, 且无需重新编译内核. eBPF提供了若干内核和用户空间事件的hook, 可以对特定系统调用, socket数据包执行操作, 如记录, 转发, 丢弃等, 因此适合用于进行性能分析, 网络处理, 系统观测等工具.

现试开发一套基于eBPF技术的容器权限监控和配置工具, 在开发阶段对容器的行为进行观测和统计, 输出报表, 并尝试为相应容器生成一个最小的系统权限配置.

## 关键成果

- 大幅安全性提升. 通过对容器行为的观测, 将容器获准的系统调用数目从默认的200+个降低至几十或十余个, 极大降低安全风险.

- 低性能开销. 基于Linux内核的`SYS_ENTER`追踪点, 实现对系统调用进入点的抽象, 相较于为400+个系统调用分别挂载追踪程序, 只需挂载1个BPF程序, 极大降低内核负载.

- 开箱即用. 用户无需了解BPF技术和编写BPF程序即可获得应用的推荐配置和行为报告, 并可直接用于容器的后续部署.

- 无内核版本依赖. 可运行在任何引入了eBPF的Linux内核上, 不依赖具体的内核版本, 即使系统调用有增删变化也能正常工作.

- 对容器透明. SecMonitor运行在目标容器外, 对容器运行不产生影响. 在此基础上, 可以从用户视角编写模拟脚本, 或直接运行已有的单元测试.

## 任务调研

### eBPF技术

见参考文献1, 2, 7, 9
- [eBPF](https://ebpf.io)
- [BCC](https://github.com/iovisor/bcc)
- [Linux Observability with BPF, by David Calavera, Lorenzo Fontana, O'Reilly](https://www.oreilly.com/library/view/linux-observability-with/9781492050193/)
- [bpf-helpers(7) — Linux manual page](https://man7.org/linux/man-pages/man7/bpf-helpers.7.html)

### 容器与cgroup技术

见参考文献3, 6, 8
- [Docker-py](https://docker-py.readthedocs.io/en/stable/)
- [Seccomp security profiles for Docker - Docker Documentation](https://docs.docker.com/engine/security/seccomp/)
- [Nginx Container - Docker Hub](https://hub.docker.com/_/nginx)

### SECCOMP技术

见参考文献6, 10
- [Seccomp security profiles for Docker - Docker Documentation](https://docs.docker.com/engine/security/seccomp/)
- [Seccomp BPF (SECure COMPuting with filters)](https://www.kernel.org/doc/html/v4.19/userspace-api/seccomp_filter.html)

## 框架设计

![arch](./archu.svg)

### 系统操作时序

为确保BPF观测到的系统调用记录无遗漏, SecMonitor确定了系统内不同操作的时序, 在架构图中用括号数字标注. 

1. 首先由Monitor向内核中挂载BPF C代码并开始监测, 获得的信息暂时在PERF Buffer中缓存. 
2. 随后Docker Client启动容器, 并利用docker SDK获取进程列表, 提供给Monitor作为filter. 
3. 容器启动后, 同时启动Emulator进程, 模拟真实环境对容器发起多种业务请求. 
4. 获得进程信息后, Monitor开始从缓存中拉取(pull)BPF记录并处理, 一段时间后观测终止, 生成中间文件(Statistics File). 
5. Analyzer最后启动, 读取中间文件, SECCOMP默认配置和系统调用映射表等必要信息, 为用户生成可读报告和新的SECCOMP配置文件.

### 模块功能描述

- Monitor: 主线程. 向内核挂载BPF程序, 读取并处理内核返回的观测信息
- Docker Client: 启动容器, 并生成进程信息, 提供给Monitor
- Emulator: 独立线程. 加载合适的业务环境模拟脚本, 向容器发起业务请求.
- Analyzer: 尾部调用. 读取系统信息, SECCOMP配置模板, 中间统计文件等, 为用户生成最终报告和新的配置文件.

## 开发计划

### 调研准备 (4.12-4.30)

- [x] BPF工具链, 了解能力范围
- [x] 阅读BCC开发文档和范例, 设计监测进程草图
- [x] 学习docker工具链和容器技术, 设计项目结构草图

### 原型开发 (5.1-5.5)

- [x] 开发第一个简单BPF C程序, 报告内核中的系统调用
- [x] 阅读kernel代码, 设计BPF C程序的挂载位置
- [x] 根据需求编写功能性BPF C程序, 使用bpftool就eBPF的监控和报告能力进行实验
- [x] 测试docker命令行工具, 就容器的管理和统计能力进行实验
- [x] 开发BCC程序, 将eBPF的挂载和信息读取自动化
- [x] 开发docker client侧程序, 将容器的启动和统计信息获取自动化
- [x] 开发Monitor主程序, 对内核信息进行筛选和统计
- [x] 编写中期报告

### 功能完善 (5.6-6.5)

- [x] 修改docker client侧程序, 利用cgroup技术实现对多进程容器的监测
- [x] 修改BPF C程序, 实现更简明的系统调用感知
- [x] 修改Monitor主程序, 实现容器行为报告文件的生成
- [x] 编写Tester进程原型, 实现较准确的容器行为观测
- [x] 编写Analyzer程序, 实现SECOMP配置文件的生成
- [x] 完成项目README文档和过程文档

### 功能扩展 (6.6-6.30)

- [ ] 编写适用于不同容器类型的测试用例, 实现更准确的容器行为观测
- [ ] 编写统计文件收集工具和分布式的SecMonitor, 实现集群上的容器行为观测
- [ ] 完善统计报告格式, 实现图形化的展示工具
- [ ] 调试工具代码, 实现生产环境稳定性

## 里程碑

- [x] (4.22-4.30) 完成相关技术调研
- [x] (4.22-4.30) 完成首个BPF C程序的挂载和运行
- [x] (5.1-5.5) 完成原型开发, 能够按需获取内核信息
- [x] (5.6-5.15) 完善原型功能, 能够和docker服务器进行交互, 监测指定容器
- [x] (5.15) 提交中期文档
- [x] (5.16-6.5) 调试程序, 完成端到端功能开发, 能够生成直接可用的报告和配置文件
- [x] (6.1-6.5) 编写项目文档

## 系统测试

针对简易nginx容器编写了测试用例, 成功观测到21种系统调用, 以json格式生成了统计报告和SECCOMP配置文件.

图片见本项目的[assets文件夹](../assets).

生成的样例报告见[results文件夹](../results).

## 问题与解决方法

- 最新的Linux x86内核中有超过400个系统调用, 如果为每个系统调用开发一段BPF检测, 内核将不堪重负, 如何解决?

通过阅读Linux内核源代码, 发现系统调用定义了一个抽象跟踪点, 所有的系统调用都会触发这一跟踪点. 将BPF挂载到此处可以获得系统调用id, 只需要一段BPF C程序就可监控全部系统调用. 参见参考文献5.

- 内核系统调用产生的频率相当高, 如何安排容器启动, 测试器启动, BPF信号处理等操作的时序, 确保系统调用记录没有因为延迟造成遗漏?

利用eBPF提供的缓存, 先启动BPF程序开始观测, 再启动容器, 最后开始读取缓存, 可以保证观测开始的时间总是不晚于容器内的首个系统调用.

- 基于动态观测的, 统计意义上的系统调用集合总是不大于实际所需的集合, 也即我们几乎总是低估实际所需的权限, 可能导致容器缺乏足够的权限来正常运行, 如何解决这一问题?

设计了独立的Emulator模块, 通过编写模拟脚本, 向容器发起业务请求, 尽可能还原生产环境, 使容器产生尽可能多的系统调用.

- 如何同时监测容器中的多个线程?

容器内的进程共享一个唯一的cgroup id. 令BPF C程序报告观测到的进程所在的cgroup id, 并以此为依据筛选记录.

- 容器中进程的id不断变化, 如何确保filter总是能够筛选出正确的调用记录?

与上题相似, 一旦获取到首条进程的信息后, 就转为依据cgroup筛选记录而非依据进程id.

- Linux的系统调用列表总在不断更新, 如何实现SecMonitor工具与内核版本的解耦合?

与问题1相似, 利用抽象跟踪点, SecMonitor的代码将与内核版本无关, 而是在运行时根据具体的内核版本进行id与函数名的映射.

- Linux的```SYS_ENTER```追踪点只提供了系统调用的id, 如何获取其与函数名的映射关系?

linux内核源代码中的`/arch/<platform-type>/entry/syscalls/`目录下存储了syscall id和名字的对应关系, 可以利用python的sys API在运行时抓取.

- 如何自动化地获取符合本机运行环境的映射表?

与上题相似, 通过读取位于本机`/usr/local/src`的内核源代码目录, 抓取系统调用表格

- docker-py SDK的运行参数非常多, 如何在避免过大编码开销的情况下确保用户能够利用docker的全部配置能力?

允许用户向monitor进程传递一个json配置文件, 运行时加载为python字典, 以`**kwargs`形式传递给Docker SDK.

## 代码目录

```
SecMonitor
├── analyzer.py
├── arch.py
├── bpf_raw_tp.c
├── default.json
├── monitor.py
├── tester.py
└── x86
    └── syscall_64.tbl
```

## 比赛收获

本次比赛使我对操作系统的功能和具体实现有了深刻的理解. 通过运用eBPF技术和阅读挖掘linux内核代码, 我对操作系统模块的具体实现有了直接的感知. 由于eBPF程序触发频率极高, 我对实时性的考量, 和系统中操作时序的安排能力也获得了较大的提升. 容器技术是云计算时代的重要技术, 本次比赛使我能够直接接触这项技术, 为后续现代操作系统领域的学习打下基础.

## 参考文献

1. [eBPF](https://ebpf.io)
2. [BCC](https://github.com/iovisor/bcc)
3. [Docker-py](https://docker-py.readthedocs.io/en/stable/)
4. [syscall_64.tbl - Linux Kernel Archive](https://git.kernel.org/pub/scm/linux/kernel/git/stable/linux.git/tree/arch/x86/entry/syscalls/syscall_64.tbl?h=v5.15.44)
5. [syscalls.h - Linux Kernel Archive](https://git.kernel.org/pub/scm/linux/kernel/git/stable/linux.git/tree/include/trace/events/syscalls.h?h=v5.15.44)
6. [Seccomp security profiles for Docker - Docker Documentation](https://docs.docker.com/engine/security/seccomp/)
7. [Linux Observability with BPF, by David Calavera, Lorenzo Fontana, O'Reilly](https://www.oreilly.com/library/view/linux-observability-with/9781492050193/)
8. [Nginx Container - Docker Hub](https://hub.docker.com/_/nginx)
9. [bpf-helpers(7) — Linux manual page](https://man7.org/linux/man-pages/man7/bpf-helpers.7.html)
10. [Seccomp BPF (SECure COMPuting with filters)](https://www.kernel.org/doc/html/v4.19/userspace-api/seccomp_filter.html)
