use std::collections::HashMap;
use std::error::Error;
use serde_derive::{Deserialize, Serialize};
use std::fs::{File, OpenOptions};
use std::hash::Hash;
use warp::reply::Response;

#[derive(Deserialize, Serialize)]
pub(crate) struct Statistics {
    data: HashMap<u32, u64>,
}

pub(crate) fn append_report_by_project(project_id: u32, stats: Statistics) -> Result<(), Box<dyn Error>>{
    let file = OpenOptions::new()
        .write(true)
        .read(true)
        .create(true)
        .open("secmonitor-stats-" + project_id.to_string() + ".json");

    let result = file.clone()?;

    let mut current_stats: HashMap<u32, u64> = serde_json::from_reader(file)?;

    for (k, v) in stats.iter() {
        match current_stats.get_mut(k) {
            Some(current_v) => { *current_v += v; },
            None => { current_stats.insert(*k, *v); },
        }
    }

    serde_json::to_writer(result, &current_stats)?;

    Ok(())
}